dnl ---------------------=|=--- pmpart.m4 ---=|=-----------------------------
dnl --=-- Protected-Mode code & data for the TUNES BIOS-based i386 API --=--
dnl
dnl --- See $TUNES/doc/info/BIOS-API.nfo for more info about this. ---
dnl
dnl
Provide({pmpart},{
dnl
dnl

pmpart


__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Orig_INT_NUM	=	0x10	; first interrupt called




;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


XtdMem:		.long	0
ConvMem:	.long	0
AvConvMem:	.long	0

CPUType:	.word	0

;;; (!) Variables a Relocator GC shall be aware of:
})


__udataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uninitialized Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (!) This must be at the very beginning,
;;; and __GDT__ must be at the very end of __pmdata__
__GDT_xxx__		= alallot(u,16,GDT_SIZE-Orig_GDT_SIZE)
__IDT__			= alallot(u,16,IDT_SIZE)
__Stack__		= alallot(u,16,PM_STACK_SIZE)
__Stack_Top__		= allot(u,0)
__Safe_Stack__		= alallot(u,16,PM_STACK_SIZE)
__Safe_Stack_Top__	= allot(u,0)
K_TSS			= alallot(u,16,sizeof_TSS32)
Dummy_TSS		= alallot(u,16,sizeof_TSS32)
V86_TSS = K_TSS
;V86_TSS 		= alallot(u,16,sizeof_TSS32+0x2000+1) ; full IO_map
User_TSS		= alallot(u,16,sizeof_TSS32)


PM_Part_Phys	= alallot(u,4)	; Physical address of PM Part
PhysicalMem	= alallot(u,4)	; address of Physical Memory in PM Part segment
RM_Part_Phys	= alallot(u,4)	; Physical address of RM Part
RM_Part_Addr	= alallot(u,4)	; address of RM Part in PM Part segment

TopFreeMem	= alallot(u,4)	; top of free memory

Init_Stream_base	= alallot(u,4)	; base of Init Stream
Init_Stream_len		= alallot(u,4)	; length of Init Stream

ReflINT_V86INT	= alallot(u,4,sizeof_V86INT)


__RM_IDT__	= alallot(u,16,4*256)


})

__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Calling Conventions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;Input:
;; *unless stated otherwise*,
;;  *if used*, the following registers are assumed to have these values:
;;   cs=KL0CS
;;   ds,es,ss=KL0DS
;;   fs=ALLMEMDS
;;   gs=RMPARTDS
;;   in EFLAGS, DF=0
;;  Assembly routines are near called
;;
;;Output:
;; *unless stated otherwise*,
;;  all segments conserved,
;;  arithmetical condition flags lost, other flags conserved.
;;  data may be used on stack
;;  all general-purpose register saved

;;  FORTH words may assume usual calling conventions
;;  FORTH words may use all their stacks
;;  FORTH words may waste their caller-save registers

;;  Interrupt handlers save everything


dnl
dnl ----- Allowing far calls to near routines -----
dnl
dnl

NearToFar:
;;Input:
;; [esp]: far return address
;; [esp+8]: near routine to call
;;Output:
;; near routine called. Far return.
	STOP(fs,0xB8000,0,14,'N,0x0B)
	call [esp+8]
	STOP(fs,0xB8000,1,14,'F,0x0B)
	retf


})


__codeseg__({

;;;;;;;;;;;;;;;;;;;;;;;;;;;;; System tables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; Descriptor tables
LibProc({SetDescriptor},edx)
LibProc({GetDescriptor},edx)



Set_INT_Vect:
;;; Set interrupt vector
;;Input:
;; bl: interrupt number
;; ebx: 32 bit offset in code
;;Output:
;; eax,edx scrambled
	mov eax,#(0x8E00<<16)|KL0CS
Set_INT_Vect_Raw:
	movzx edx,bl
	lea edx,[edx*8+pmaddr(__IDT__)]
SetGate:
;;Input:
;; edx points to the descriptor to set
;; ebx is the offset in the descriptor for call/interrupt/trap gates
;;     must be zero for task gates
;; eax contains:
;;   bits 00-15=selector pointed by the gate
;;   bits 16-23=counter for call gates, 0 for other gates
;;   bits 24-31=P,DPL,Type
;;Output:
;;   descriptor set, no register lost
	mov [edx],ebx
	mov [edx+4],ebx
	mov [edx+2],eax
	ret


Get_INT_Vect:
;;Input:
;; bl: interrupt number
;;Output:
;; eax,edx scrambled
;; ebx: 32 bit offset of PM interrupt handler in KL0CS
	movzx edx,bl
	lea edx,[edx*8+pmaddr(__IDT__)]
GetGate:
;;Input:
;; edx points to the descriptor to set
;;Output:
;; ebx is the offset in the descriptor for call/interrupt/trap gates
;;     must be zero for task gates
;; eax contains:
;;   bits 00-15=selector pointed by the gate
;;   bits 16-23=counter for call gates, 0 for other gates
;;   bits 24-31=P,DPL,Type
	mov eax,[edx+2]
	mov ebx,[edx+4]
	mov bx,[edx]
	ret


Init_TSS:
;;Input:
;; [edi]: TSS32 structure
;; ebx: target EIP
;; eax: target CS (DS,ES,SS must be CS+8)
;; esi: target ESP
;; ecx: target EFLAGS
;;Output:
;;; Initialize the TSS structure
;; all gregisters conserved
   ;; Set registers immediately given:
	mov [edi+TSS32_EFLAGS],ecx
	mov [edi+TSS32_ESP],esi
	mov [edi+TSS32_EIP],ebx
	mov [edi+TSS32_CS],eax
   ;; Set data registers to CS+8
	add eax,#8
	mov [edi+TSS32_DS],eax
	mov [edi+TSS32_ES],eax
	mov [edi+TSS32_SS],eax
	sub eax,#8

Init_TSS0:
;;Input:
;; [edi]: TSS32 structure
;;Output:
;;; Initialize the TSS structure
	push eax
   ;; Page table (?)
	ttmov([edi+TSS32_CR3],[pmaddr(K_CR3)],eax)
   ;; Constant entries
	mov [edi+TSS32_ESP_CPL0],#pmaddr(__Safe_Stack_Top__)
	mov [edi+TSS32_TrapBit],#(sizeof_TSS32<<16)
	clr(eax)
	mov [edi+TSS32_BackLinkSelector],eax
	mov [edi+TSS32_ESP_CPL1],eax
	mov [edi+TSS32_ESP_CPL2],eax
	mov [edi+TSS32_LDT],eax
	mov al,#ALLMEMDS
	mov [edi+TSS32_FS],eax
	mov al,#KL0DS
	mov [edi+TSS32_GS],eax
	mov [edi+TSS32_SS_CPL0],eax
	mov al,#KL1DS
	mov [edi+TSS32_SS_CPL1],eax
	mov al,#KL2DS
	mov [edi+TSS32_SS_CPL2],eax
	pop eax
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; I/O subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


LibProc(Disable_NMI)		; al scrambled
LibProc(Enable_NMI)		; al scrambled

Mask_IRQs:
;;Output:
;;; all IRQs masked
;; al scrambled
	outb(0xA1,#0xFF)	; mask off all slave irq's for now
	outb(0x21,#0xFB)	; mask off all master irq's but irq2 cascade
	ret

Unmask_IRQs:
;;Output:
;;; all IRQs unmasked
;; al scrambled
	outb(0x21,#0x00)	; unmask all master irq's
	outb(0xA1,#0x00)	; unmask all slave irq's
	ret

Setup_PM_8259:
;;Output:
;;; 8259 PIC reprogrammed for PM
;; al,cx scrambled
	mov cl,#IRQ_INT_BASE
	mov ch,#IRQ_INT_BASE+8
	jmp Setup_8259

Setup_RM_8259:
;;Output:
;; al,cx scrambled
;; 8259 PIC reprogrammed for RM
	mov cl,#0x08
	mov ch,#0x70
	;jmp Setup_8259

Setup_8259:
;;Input:
;; cl: base of interrupts for IRQ0-7
;; ch: base of interrupts for IRQ8-F
;;Output:
;; al scrambled
;; 8259 PIC reprogrammed with IRQs at right interrupt
    ;;; Programming 8259A-1 (Master)
	outb(0x20,#0x11)		; initialization sequence
	outb(0x21,cl)			; start of master hardware int's
	outb(0x21,#0x4)			; 8259-1 is master
	outb(0x21,#0x1)			; 8259-1 in 8086 mode
    ;;; Programming 8259A-2 (Slave)
	outb(0xA0,#0x11)		; initialization sequence
	outb(0xA1,ch)			; start of slave hardware int's
	outb(0xA1,#0x2)			; 8259-2 is slave
	outb(0xA1,#0x1)			; 8259-2 in 8086 mode
    ;;; Done
	ret

Flush_Interrupts:
;;; Flush possible pending (possibly keyboard) interrupt
;;; Many thanks to Adam Seychell about that !
;;Output:
;; al scrambled
;; EOI sent to both PICs
;; character read from keyboard controller interface
	mov al,#0x20
        out #0x20,al	; send EOI to PIC1
        out #0xA0,al	; send EOI to PIC2
	in al,#0x60	; read keyboard
	ret

Reset_Copro:
	clr(ax)
	out 0xF0,al
	call Delay
	out 0xF1,al
	ret




;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uncaught Interrupts ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


Uncaught_Exception:
;;; Procedure called to show that an Uncaught Exception occurred.
;;; It show the error, and
;;Input:
;; [esp]: pushad,exception frame
;; dl: exception to report
;;Output:
;; interrupted machine state showed,
;; execution halted
;;To Do:
;; have a hook for that, and have video-mode specific reporting (?)
	pushm(es,ds,fs,gs)
	ppmov({ds,es},{ss,ss})
	ppmov(fs,#ALLMEMDS)
	STOP(fs,0xB8000,6,0,'X,0x0F)
   ;;; Some V86 reflection
	test byte [esp+INTSTACK_D_EFLAGS+2],#2
	bne V86_Reflector
   ;;; Show that an error occurred
	cld
	mov edi,[pmaddr(Video_mem_base)]
	mov [edi],#0x8F458F55  ; blinking UE
	add edi,#4
	mov ah,#0x0F
	ppmov(ecx,#2)
	call DumpHex
   ;;; Dump interrupted registers
	mov dword [pmaddr(DumpLine)],#1
	call DumpINTSTACK
	mov esi,[esp+INTSTACK_D_EIP]
	call ShowBZone
	add esp,#sizeof_INTSTACK_D-8
	mov esi,esp
	push cs
	push esi
	sub esi,#8
	call ShowDZone
	___die___ ; Uncaught Exception !


Uncaught_IRQ:
;;; Procedure called to show that an Uncaught IRQ occurred.
;;; It show the error, and
;;Input:
;; [esp]: pushad,exception frame
;; dl: exception to report
;;Output:
;; interrupted machine state showed,
;; execution halted
;;To Do:
;; have a hook for that, and have video-mode specific reporting (?)
   ;;; Show that an error occurred
	pushm(es,ds,fs,gs)
Uncaught_IRQ__:
	ppmov(ds,ss)
	ppmov(es,ds)
	ppmov(fs,#ALLMEMDS)
	cld
	mov edi,[pmaddr(Video_mem_base)]
	mov [edi],#0x8F538F49  ; blinking IR
	add edi,#4
	mov ah,#0x0F
	mov ecx,#2
	call DumpHex
   ;;; Dump interrupted registers
	mov dword [pmaddr(DumpLine)],#1
	call DumpINTSTACK
	___die___ ; Uncaught IRQ !


;;;;;;;;;;;;;;;;;;;;;;;;;;;; Exception Handling ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


Division_Error_Trap:
	push #0
	pushad
	mov dl,#0x0
	br Uncaught_Exception

Single_Step_Trap:
	push #0
	pushad
	mov dl,#0x1
	br Uncaught_Exception

NMI_Trap:
	push #0
	pushad
	mov dl,#0x3
	br Uncaught_Exception

BreakPoint_Trap:
	push #0
	pushad
	mov dl,#0x3
	br Uncaught_Exception

Overflow_Trap:
	push #0
	pushad
	mov dl,#0x4
	br Uncaught_Exception

Out_of_Bound_Trap:
	push #0
	pushad
	mov dl,#0x5
	br Uncaught_Exception

Invalid_Opcode_Trap:
	push #0
	pushad
	mov dl,#0x6
	br Uncaught_Exception

Emulate_Coproc_Trap:
	push #0
	pushad
	mov dl,#0x7
	br Uncaught_Exception

Double_Fault_Stop:
;; an error code is pushed
	pushad
	mov dl,#0x8
	br Uncaught_Exception

Coproc_Segment_Stop:
	push #0
	pushad
	mov dl,#0x9
	br Uncaught_Exception

Invalid_TSS_Fault:
;; an error code is pushed
	pushad
	mov dl,#0xA
	br Uncaught_Exception

Segment_Not_Present_Fault:
;; an error code is pushed
	pushad
	mov dl,#0xB
	br Uncaught_Exception

Stack_Fault:
;; an error code is pushed
	pushad
	mov dl,#0xC
	br Uncaught_Exception


General_Protection_Fault:
;;; General protection violation Handler
;;Input:
;; frame indicating the machine state interrupted by a general protection fault
;; an error code is pushed
;;Output:
;;Done:
;; giving hand to the V86 GPF handler if needs be.
;; else report an uncaught exception
;;To Do:
;; almost everything
    ;; is interrupted routine's FLAG_VM bit set ?
	test byte [esp+4+INTSTACK_EFLAGS+2],#2 ; +4:skip error code
	bne V86_GP_Handler
	zzzz({
	   pushing({fs},{
	   ppmov(fs,#ALLMEMDS)
	   STOP(fs,0xB8000,18,0,'G,0x0F)
	   STOP(fs,0xB8000,19,0,'P,0x0F)
	})})
        pushad
	mov dl,#0xD
	br Uncaught_Exception


Page_Fault:
;; an error code is pushed
	pushad
	mov dl,#0xE
	br Uncaught_Exception

Reserved_0F_Fault:
	push #0
	pushad
	mov dl,#0xF
	br Uncaught_Exception

Coprocessor_Error_Fault:
	push #0
	pushad
	mov dl,#0x10
	br Uncaught_Exception

Alignment_Check_Fault:
;; an error code is pushed
	pushad
	mov dl,#0x11
	br Uncaught_Exception




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; IRQ Handling ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Reflect_IRQ_to_RM:
;;Input:
;; cl: interrupt number to reflect
;; ds,es: any
;;Output:
;; everything scrambled but ss:esp
;; RM IRQ handler called in current X86 mode
;;To Do:
;; verify reentrance
	ppmov(ds,ss)
	ppmov(es,ds)
	mov eax,[pmaddr(X86_FLAGS_Addr)]
	pushing({eax,[eax],[pmaddr(Current_V86INT)]},{
	mov ebx,#pmaddr(ReflINT_V86INT)
	mov [pmaddr(Current_V86INT)],ebx
	call Call_INT_R86
	})
	ret

define({DefReflIRQ},
{
ReflIRQ_$2:
	pushing({es,ds,fs,gs,ad},{
	ppmov(fs,#ALLMEMDS)
	STOP(fs,0xB8000,(64+0x$2),0,'$2,0xE1)
	mov cl,#$1+0x$2
	call Reflect_IRQ_to_RM
	STOP(fs,0xB8000,(64+0x$2),0,'$2,0x61)
	})
	iretd
})



IRQ0_Timer:
	DefReflIRQ(0x08-0,0)

IRQ1_Keyboard:
	DefReflIRQ(0x08-0,1)

IRQ2_Cascade:
	DefReflIRQ(0x08-0,2)

IRQ3:
	DefReflIRQ(0x08-0,3)

IRQ4:
	DefReflIRQ(0x08-0,4)

IRQ5:
	DefReflIRQ(0x08-0,5)

IRQ6:
	DefReflIRQ(0x08-0,6)

IRQ7:
	DefReflIRQ(0x08-0,7)

IRQ8_RTC:
	DefReflIRQ(0x70-8,8)

IRQ9:
	DefReflIRQ(0x70-8,9)

IRQ10:
	DefReflIRQ(0x70-8,A)

IRQ11:
	DefReflIRQ(0x70-8,B)

IRQ12:
	DefReflIRQ(0x70-8,C)

IRQ13:
	DefReflIRQ(0x70-8,D)

IRQ14:
	DefReflIRQ(0x70-8,E)

IRQ15:
	DefReflIRQ(0x70-8,F)





})

})dnl --- end of pmpart.m4
dnl
