dnl ---------------------=|=--- rmpart.m4 ---=|=-----------------------------
dnl --=-- Real-Mode code & data for the TUNES BIOS-based i386 API --=--
dnl
dnl --- See ~moose/info/BIOS-API.nfo for more info about this. ---
dnl
dnl
Provide({rmpart},{
dnl
dnl



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


rmpart

__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


PM_Target:		; Address to jump to when entering PM
	PM_Target_ofs:	.long	pmaddr(PM_Entry_Point)
	PM_Target_seg:	.word	PMPARTCS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PM_GDT_address:		; Length & Base address of protected-mode GDT
	PM_GDT_lim:	.word	GDT_SIZE-1
	PM_GDT_base:	.long	Orig_GDT

})



__udataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uninitialized Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

alignheap(ru,4)
; (!) StackTop mustn't be at the end, or it will be 64K, and
; esp=0x10000 is very bug-prone when going RM to PM !!!!
RMPartStackBot	= allot(ru,256*2)
RMPartStackTop	= allot(ru,0)

RM_cr0		= allot(ru,4)	; value of cr0 for entering real mode
PM_cr0		= allot(ru,4)	; value of cr0 for entering protected mode

;RM_STACK 	= alallot(ru,4,0) 	; Saved addr for R-mode stack
;RM_STACK_ofs	= allot(ru,2)
;RM_STACK_seg:	= allot(ru,2)

X86_FLAGS	= allot(ru,4)
X86_XisV	= 0x01

})


__codeseg__({
pushdef({DumpERegs},{RM_DumpERegs})dnl

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; I/O subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pushdef({__IO_Delay__},{call Delay})dnl

;;; Delay between I/O, as in Linux.
Delay:
	jmp DelayRet
DelayRet:
	ret

RM_Mask_IRQs:
	outb(0xA1,#0xFF)	; mask off all interrupts for now
	outb(0x21,#0xFB)	; mask all irq's but irq2 which is cascaded
	ret

RM_Unmask_IRQs:
	outb(0x21,#0x00)
	outb(0xA1,#0x00)
	ret

RM_Disable_NMI:
	LibInline(Disable_NMI)
	ret

RM_Enable_NMI:
	LibInline(Enable_NMI)
	ret

LibProc(KB_Ready,Timeout_8042)

; What to do in case of time out from 8042
Timeout_8042:
	jmp KB_Ready

Disable_A20:
	mov ah,#set_A20_OFF
	jmp Set_A20

Enable_A20:
	mov ah,#set_A20_ON
	;jmp Set_A20

LibProc(Set_A20)



popdef({__IO_Delay__})





;;;;;;;;;;;;;;;;;;;;;;;;;;; Support for V86 Mode ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

V86_REQUEST:
;;Input:
;; [sp+4]: X86RQ request number
;; [sp]: far return address
	int #V86_SERVER_INT
	retf 2

X86_RETURN:
;;Input:
;; [sp]: far return address
	push #X86RQ_RETURN

X86_IMM_REQUEST:
	pushing({bp},{
	mov bp,sp
	ppmov({dword [bp+2],word [bp+6]},{dword [bp+4],word [bp+2]})
	})

X86_REQUEST:
;;Input:
;; [sp+4]: X86RQ request number
;; [sp]: far return address
;;Output:
;; call the right routine of R86_REQUEST and V86_REQUEST
	seg cs
	test [rmaddr(X86_FLAGS)],#X86_XisV
	jnz V86_REQUEST
;	jmp R86_REQUEST


;;;;;;;;;;;;;;;;;;;;;;;;;;; Support for R86 Mode ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;; Save State for R86 ;;;;;

R86_REQUEST:
Save_State:
;;; Save R86 machine state and go PM
;;Input:
;; [sp+4]: X86RQ request
;; [sp]: cs:ip to save
;;Output:
;; * a V86INT frame is created on stack,
;; * execution is given to protected mode.
	push bp
	push ax
	clr(ax)
	mov bp,sp
	pushm(fd,esp,ss,ax,es,ax,ds,ax,fs,ax,gs,ax)
	pushm([bp+4],ax,[bp+6],ax) ; return CS:IP
	add [bp-6*4],#5*4+5*2 ; adjust esp for ss,es,ds,fs,gs;ax,bp,ip,cs,RQ
	mov ax,[bp]
	mov bp,[bp+2]
	pushad
	STOP(fs,0xB8000,41,24,'S,0x0B)
	;jmp GoProtect

;;;;; Real Mode to Protected Mode switcher ;;;;;
GoProtect:
;;Input:
;; cs: rmpart segment
;;Output:
;; * A20 line is enabled
;; * all hardware interrupts are disabled
;; * Processor enters Protected mode
;; * Processor jumps into 32-bit far address contained in [rmaddr(PM_Target)]
;; * eax,ebx,ecx,edx scrambled
;; * ds moved to RM_Part_Seg
;; * es,ss,fs,gs unchanged (internal values also)
;; * esp,ebp,esi,edi unchanged

   ;zwoz({{"Going PM, Target="}},[rmaddr(PM_Target_seg)],{":"})
   ;zdwoz({},[rmaddr(PM_Target_ofs)],{CRLF})
	cli
   ;;__("Disabling NMI",CRLF)
	;;call RM_Disable_NMI
   ;__("Enabling A20",CRLF)
	call Enable_A20
	zzzzz({call WaitGoPM})
   ;zdwoz({"Loading GDT at "},[rmaddr(PM_GDT_base)],{CRLF})
	ppmov(ds,cs)
	lgdt [rmaddr(PM_GDT_address)]
	mov eax,[rmaddr(PM_cr0)]	; 32 bit msw
	test byte [rmaddr(GP_CR0)],#1
	mov cr0,eax			; This is it!
 GP_CR0:
	jmpi dword [rmaddr(PM_Target)]


;;;;; Protected to Real Mode switcher ;;;;;

GoReal:
;;Input:
;; * Interrupts are disabled,
;; cs: pmpart segment
;;Output:
;; eax scrambled
;; * (just) do the switching to Real mode and jump to 16-bit far [RM_Target]
;;NOTE:
;; I had problems having this code in a 32 bit segment,
;; as the final jump seemed to get nowhere (or get it in real 32-bit mode ?)
	zzzz({
	  mov ax,#ALLMEMDS
	  mov fs,ax
	})
	STOP(fs,0xB8000,3,22,'G,0x0F)
	STOP(fs,0xB8000,1,15,'G,0x0C)
	seg cs
	mov eax,[rmaddr(RM_cr0)]	; 16 bit real-mode msw
	STOP(fs,0xB8000,2,15,'o,0xC)
	zzzz({call WaitGoRM})
	seg cs
	test byte [rmaddr(GR_CR0)],#1 ; put following instr. in cache...
	mov cr0,eax		; Back to the Real World :( ):
   GR_CR0:
	flushinstr
	STOP(fs,0xB8000,3,15,'R,0xC)

	.byte 0xEA	;seg cs	;jmpi word [RM_Target]
RM_Target:
RM_Target_ofs:	.word	rmaddr(Restore_State)
RM_Target_seg:	.word	0 ; ??? Will be [RM_Part_seg]


;;;;; Restore State from R86 ;;;;;
;;Input:
;; [sp]: the zero-padded values for es,ds,fs,gs
;; [sp+16]: a pushad/popad frame
;; [sp+48]: a iretd frame (eip,cs,0,fd)
;;Output:
;; * Restore_State_fg restores everything
;; * Restore_State restores everything but fs,gs
;;  with typical use fs=ALLMEM and gs=PMPART

Restore_State_fg:
	popm(ad,dword es,dword ds,dword fs,dword gs)
	lss esp,[esp]
	iretd

Restore_State:
	STOP(fs,0xB8000,0,0,'R,0x61)
	zzzz({
	  seg cs
	  mov word [rmaddr(RM_DumpLine)],#18
	})
	STOP(fs,0xB8000,1,0,'S,0x61)
	popm(ad,es,ds,ds)
	STOP(fs,0xB8000,2,0,'8,0x61)
	;zWZonez(,[esp+4+10+sizeof_PUSHA+2],[esp+10+sizeof_PUSHA+2],16,{CRLF})
	STOP(fs,0xB8000,3,0,'6,0x61)
	lss esp,[esp+10]
	STOP(fs,0xB8000,4,0,'D,0x61)
	;rz__z
	STOP(fs,0xB8000,5,0,'Z,0x61)
	;___die___
	iretd


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Support for interrupts ;;;;;;;;;;;;;;;;;;;;;;;;

RM_Idle:
;;; indefinite idle loop to wait for interrupts
	;cli
	;hlt	; idle loop for interrupt reflectors.
	jmp RM_Idle




;;; Hook from hardware IRQ to BIOS handlers
;;; (when TUNES I/O control is disabled)

RM_IRQ:

define({DefEmulIRQ},
{RM_IRQ_$2:
	STOP(fs,0xB8000,(48+0x$2),0,'$2,0x8F)
	int #$1+0x$2
	STOP(fs,0xB8000,(48+0x$2),0,'$2,0x0F)
	iret
})

irp(X,{0,1,2,3,4,5,6,7},{DefEmulIRQ(0x08-0,X)})
irp(X,{8,9,A,B,C,D,E,F},{DefEmulIRQ(0x70-8,X)})
RM_IRQ_len = RM_IRQ_1 - RM_IRQ_0


})


;;; Reflector for RM to PM interrupts

; one way to do things is to intercept everything through our one
; RM IDT outside of conventional memory which points to this routine,
; with low byte of CS being the interrupt number,
; except perhaps for hardware interrupts.

INT_Interceptor:
	push cs
	.byte 0xEA ; jmpi
Actual_INT_Interceptor:
	.word rmaddr(INT_Interceptor_0)
Actual_INT_Interceptor_seg:
	.word 0 ; ?

INT_Interceptor_0:
	push #X86RQ_RM_INT
	push cs
	call R86_REQUEST
	iret





;;;;;;;;;;;;;;;; Calling 32-bit routines from 16-bit PM ;;;;;;;;;;;;;;;;;;;;;

define({CallPM32},
       {push dword #pmaddr($1)
	call Call_PM32})

Call_PM32:
;;Input:
;; * CPU in 16-bit protected mode
;; [esp+2]: 32-bit near address to call in KL0CS
;; [esp]: 16-bit near return address
;;Output:
;; 32-bit routine called
	push dword #KL0CS
	push dword #pmaddr(NearToFar)
	push dword [esp+10]
	calli dword [esp+4]
	lea esp,[esp+12]
	ret 4





popdef({DumpERegs})dnl

})dnl --- end of rmpart.m4
dnl
