dnl ---------------------=|=--- xrv86.m4 ---=|=-----------------------------
dnl --=-- Real-Mode and V86 mode support for TUNES BIOS-based i386 API --=--
dnl
dnl R86 refers to going back to real-mode
dnl V86 refers to emulating real-mode using the V86 mode of 386+ CPUs
dnl X86 refers to either of the previous.
dnl
dnl --- See $TUNES/doc/info/BIOS-API.nfo for more info about this. ---
dnl
Provide({xrv86},{
dnl
dnl

pmpart

__dataseg__({

RM_Part_seg:	.long	0 ; ?	; segment for RM part (long: ease of use)


; Saved address for P-mode stack
PM_STACK:
PM_STACK_ofs:	.long	pmaddr(__Stack_Top__)
PM_STACK_seg:	.word	PMPARTDS

RM_IDT_address:
RM_IDT_lim:	.word	256*4-1		; Limit of real-mode IDT
RM_IDT_base:	.long	0		; Base Address of real-mode IDT

PM_IDT_address:
PM_IDT_lim:	.word	IDT_SIZE-1	; Limit of protected-mode IDT
PM_IDT_base:	.long	pmaddr(__IDT__)	; Base Address of protected-mode IDT
					; (!) needs be adjusted
.align 4

X86_FLAGS_Addr:	.long	rmaddr(X86_FLAGS)	; (to adjust)


;;; (!) Variables a Relocator GC shall be aware of:

; Saved address for P-mode stack
Current_V86INT:
		.long	pmaddr(ReflINT_V86INT)

})

__codeseg__({
;;;;;;;;;;;;;;;;; Calling the currently active X86 V86 Handler ;;;;;;;;;;;;;;;

Call_X86:
;;Input:
;; [ebx]: V86INT structure
;;Output:
;; real-mode routine called in current X86 mode
;; and parameter struct updated
	mov eax,[pmaddr(X86_FLAGS_Addr)]
	test byte [eax],#X86_XisV
	bne Call_V86
	;jmp Call_R86 ; (!) must follow immediately



;;;;;;;;;;;;;;;;;;;;;;;;;;;; Simple R86 Handler ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; mode R86 is to call back between PM and Real Real Mode, *not* V86
; another difference is that R86 is invoked with fs=ALLMEMDS and gs=PMPARTDS
; though nobody prevents the RM code to trash fs and gs
; current version inspired by V86 code


Call_R86:
;;Input:
;; [ebx]: V86INT structure
;;Output:
;; all registers preserved
;; [ebx]: updated V86INT structure
	DMSG({},0,24,0x03,fs,0xB8000)
	DBG(C)

   ;; Save PM/LLL context
	pushm(ad,fs,gs,fd) ; ds,es assumed to be KL0DS or RMPARTDS
	DBG(a)

   ;; Begin critical section
	cli
	;call Disable_NMI
	call Disable_Paging
	DBG(l)

   ;; Save PM stack address
	mov [pmaddr(PM_STACK_ofs)],esp
	mov [pmaddr(PM_STACK_seg)],ss
	DBG(l)

   ;; Set X86_FLAGS
	mov eax,[pmaddr(X86_FLAGS_Addr)]
	and [eax],#~X86_XisV
	DBG(_)

R86_AGAIN:
	DBG(R86)
   ;; Compute and push (adjusted) RM stack address
	mov ecx,[ebx+V86INT_ESP]
	mov eax,[ebx+V86INT_SS]
	sub cx,#12
	push eax
	push ecx
	clr(edi)
	mov di,cx
	shl eax,#4
	add edi,eax
	add edi,[pmaddr(PhysicalMem)]
	DBG({ S})

   ;; "Push" EIP,CS,FD on RM stack
	ttmov([edi],  [ebx+V86INT_EIP],eax)
	ttmov([edi+4],[ebx+V86INT_CS],eax)
	ttmov([edi+8],[ebx+V86INT_EFLAGS],eax)
	DBG(P)

   ;; Extract Restore_State parameters from V86INT structure
	sub esp,#sizeof_PUSHAD+4*4
	mov edi,esp
	mov esi,ebx
	mov ecx,#8
	rep
	movsd
	add esi,#V86INT_ES-V86INT_EIP
	rept(4,{
	   movsd})
	DBG(V2R)

   ;; initialize fs,gs to ALLMEM, PMPART
	ppmov(eax,eax)
	DBG(A)
	movzil(a,#ALLMEMDS)
	DBG(1)
	mov fs,ax
	DBG(F)

	ppmov(fs,#ALLMEMDS)
	DBG(F)
	ppmov(gs,#PMPARTDS)
	DBG(G)

   ;; setup RM interrupts
	lidt [pmaddr(RM_IDT_address)]
	DBG(I)
	;movzx eax,word [pmaddr(RM_IDT_lim)]
	;mov ebx,dword [pmaddr(RM_IDT_base)]
	;z__z

   ;; Jump into R86 mode
	jmpi #rmaddr(GoReal), #RMPARTCS

Return_From_R86:
   ;; Initialize ds,es to KL0DS
	ppmov(ds,#KL0DS)
	ppmov(es,ds)
	zzzz({ppmov(fs,#ALLMEMDS)})
	STOP(fs,0xB8000,0,23,'R,0x0B)
   ;; Compute logical address of RM stack
	clr(ebx)	; bx is also cleared for below
	clr(esi)
	mov bx,sp
	mov si,ss
	shl esi,#4
	add esi,ebx
	add esi,[pmaddr(PhysicalMem)]

	STOP(fs,0xB8000,1,23,'e,0x0B)
	STOP(fs,0xB8000,2,23,'t,0x0B)
	zzzzz({mov [pmaddr(DumpLine)],#1})
	STOP(fs,0xB8000,3,23,'u,0x0B)

   ;; Restore PM stack
	lss esp,[pmaddr(PM_STACK)]
	STOP(fs,0xB8000,4,23,'r,0x0B)
	STOP(fs,0xB8000,5,23,'n,0x0B)
	STOP(fs,0xB8000,6,23,'_,0x0B)

   ;; Copy RM stack frame to parameter V86INT structure
	mov edi,[esp+PUSHAD_EBX]
	mov ebx,edi

	zzzzz({call DumpV86INT})
	STOP(fs,0xB8000,7,23,'F,0x0B)

	mov ecx,#sizeof_V86INT/4
	cld
	rep
	movsd
	STOP(fs,0xB8000,8,23,'r,0x0B)

	zzzzz({call DumpV86INT})
	STOP(fs,0xB8000,9,23,'o,0x0B)
	;z__z

   ;; End Critical Section
	;lidt [pmaddr(PM_IDT_address)]	; necessary if not paging
	call Enable_Paging
	;call Enable_NMI
	STOP(fs,0xB8000,10,23,'m,0x0B)

   ;; Get REQUEST parameter
	add esi,#8
	;mov ebx,[...] ; (!) done above
	STOP(fs,0xB8000,11,23,'_,0x0B)
	push #pmaddr(R86_AGAIN)
	call X86_SERVER
	STOP(fs,0xB8000,12,23,'R,0x0B)

R86_Do_Return:
   ;; Restore PM/LLL state
	STOP(fs,0xB8000,13,23,'8,0x0B)
	popm(ad,fs,gs,fd)
	;z__z
	STOP(fs,0xB8000,14,23,'6,0x0B)
   ;; Now we can continue PM/LLL computations
	ret



;;;;;;;;;;;;;;;;;;;;;;;;;;;; Simple V86 Handler ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; Ideas (and small bits of code) stolen from (in decreasing order of interest):
; - OOF/VMI by Andras Zsoter,
; - START32 and PMODE by Tran,
; - 386POWER by Lorenzo Micheletto,
; - Linux by Linus Torvalds & Co.



Call_V86:
;;Input:
;; [ebx]: V86INT structure
;;Output:
;; [ebx]: updated V86INT structure
;;; at a given moment, the V86 stuff is active for only one
;;;
;;;
   ;; Save PM/LLL state
	z__z
	pushm([pmaddr(V86_TSS+TSS32_ESP_CPL0)],
		 ad,fs,gs,fd)
		;was also [pmaddr(V86_TSS+TSS32_SS_CPL0)],

   ;; Critical section
	cli
	STOP(fs,0xB8000,0,18,'C,0x0F) ; 'G
   ;; Setup stack for V86 supervisor as current stack
	mov [pmaddr(V86_TSS+TSS32_ESP_CPL0)],esp
	mov [pmaddr(V86_TSS+TSS32_SS_CPL0)],ss
	zzzz({lss esp,[pmaddr(V86_TSS+TSS32_ESP_CPL0)]})
	STOP(fs,0xB8000,1,18,'a,0x0F)
   ;; Load task register
	and byte [pmaddr(__GDT__)+V86TSS+5],#~2
	movzil(a,#V86TSS)
	ltr ax
   ;; Set X86_FLAGS
	mov eax,[pmaddr(X86_FLAGS_Addr)]
	or [eax],#X86_XisV
   ;; Build a stack frame from the parameter V86INT structure
	sub esp,#sizeof_V86INT
	mov ecx,#sizeof_V86INT/4
	mov esi,ebx
	mov edi,esp
	rep
	movsd
	or byte [esp+V86INT_EFLAGS+2],#2 ; set VM bit
	STOP(fs,0xB8000,2,18,'l,0x0F)
	mov ebx,esp
	zzzzz({
	  call DumpV86INT
	})
	STOP(fs,0xB8000,3,18,'l,0x0F)
   ;; Jump into V86 !!!
	STOP(fs,0xB8000,4,18,'_,0x0F)
	STOP(fs,0xB8000,5,18,'V,0x0F)
	STOP(fs,0xB8000,6,18,'8,0x0F)
	ppmov(fd,#0x3000)
	STOP(fs,0xB8000,7,18,'6,0x0F)
	popad
	iretd


;;; Supervisor Services for V86 box
V86_SERVER:
;;Input:
;; [esi]: stack of interrupted V86INT task
;; [esp]: V86INT stack frame from interruption
	STOP(fs,0xB8000,24,0,'V,0x71)
	mov ebx,esp
	zzz({
	  SHOWL("V86_SERVER")
	  z__z
	  call DumpV86INT
	})
	push #pmaddr(V86_AGAIN)
	call X86_SERVER

   ;; Back we are !!!
Return_From_V86:
; it *seems* that there is a pushad frame too many when we reach this point !
; but the same pushad as
	zzzz({ppmov(fs,#ALLMEMDS)})
	DMSG({Return_From_V86},0,19,0x0F,fs,0xB8000)
	LETTER(4)
   ;; Initialize ds and es to KL0DS
	ppmov(ds,#KL0DS)
	ppmov(es,ds)
	LETTER(4)
   ;; Pop the V86INT stack frame into the V86INT structure
	mov edi,[esp+sizeof_V86INT+PUSHAD_EBX]
	zzz({
	  mov ebx,edi
	})
	LETTER
	mov esi,esp
	mov ecx,sizeof_V86INT/4
	cld
	rep
	movsd
	mov esp,esi
	STOP(fs,0xB8000,9,19,'o,0x0F)
	zzzzz({
	  call DumpV86INT
	})
	LETTER
   ;; Restore PM/LLL state
	LETTER(2)
	zzzz({
	  push esp
	  mov esi,esp
	  call ShowDZone
	  call ShowDZone
	})
	popm([pmaddr(V86_TSS+TSS32_ESP_CPL0)],
		 ad,fs,gs,fd)
	LETTER(2)
	zzzz({
	  push eax
	  mov eax,[esp+4]
	  z__z
	  pop eax
	})
   ;; Now we can continue PM/LLL computations
	ret



V86_Reflector:
;;; Reflector for uncaught exceptions
   ;; Canonicalize stack & get exception number
	add esp,#4*4
	STOP(fs,0xB8000,4,0,'R,0x0F)
	mov [esp+sizeof_PUSHAD+4+INTSTACK_CS+2],dl
	popad
	add esp,4
	pushad
	clr(dl)
	xchg dl,[esp+sizeof_PUSHAD+INTSTACK_CS+2]

   ;; Initialize segment registers
	ppmov(ds,ss)	; edit linear memory - (assuming Paging)
	ppmov(es,ds)
	STOP(ds,0xB8000,5,0,'R,0x0F)

   ;; Have esi point to the CS_IP linear address, while ebp contains the SP
	SegOfsToLinear(esi,bp,[esp+V86INT_ESP],[esp+V86INT_SS])

   ;; Do Reflect interrupt
	br V86_INT_n



;;; General protection violation handler for V86 mode
V86_GP_Handler:
;;Input:
;;; assuming paging is on.
;; [esp]: stack layout from a GPF that just happened from V86 mode.
;; ss=KL0DS
;;Output:
;;; treats the interrupt
   ;; Canonicalize stack frame into V86INT structure
	add esp,#4	; remove error code
	pushad		; push registers

   ;; Initialize segment registers
	ppmov(ds,ss)	; edit linear memory - (assuming Paging)
	ppmov(es,ds)

	STOP(ds,0xB8000,20,0,'V,0x1F)
	STOP(ds,0xB8000,21,0,'G,0x1F)

	zzz({ppmov(fs,ds)})
	zzzzz({
	  STOP(ds,0xB8000,21,0,'h,0x1F)
	  mov ebx,esp
	  STOP(ds,0xB8000,21,0,'i,0x1F)
	  STOP(ds,0xB8000,21,0,'j,0x1F)
	  call DumpV86INT
	  STOP(ds,0xB8000,21,0,'k,0x1F)
	})
	STOP(ds,0xB8000,22,0,'e,0x1F)

   ;; Have ebx point to the CS_IP linear address, while ecx contains the IP
   ;; Have esi point to the CS_IP linear address, while ebp contains the SP
	SegOfsToLinear(ebx,cx,[esp+V86INT_EIP],[esp+V86INT_CS])
	STOP(ds,0xB8000,22,0,'T,0x1F)
	SegOfsToLinear(esi,bp,[esp+V86INT_ESP],[esp+V86INT_SS])
	STOP(ds,0xB8000,22,0,'L,0x1F)

   ;; Read opcode of interrupted instruction
	inc word [esp+V86INT_EIP]
	mov al,[ebx]

   ;; Is it an INT3 ?
	STOP(ds,0xB8000,23,0,'3,0x1F)
	mov dl,#3
	cmp al,#0xCC
	je V86_INT_n
   ;; Is it an INTO ?
	STOP(ds,0xB8000,23,0,'4,0x1F)
	mov dl,#4
	cmp al,#0xCE
	je V86_INT_n
   ;; Is it an INT xx ?
	STOP(ds,0xB8000,23,0,'x,0x1F)
	cmp al,#0xCD
	je V86_INT
   ;; Last case: uncaught exception :-(  )-8
	STOP(ds,0xB8000,23,0,'U,0x1F)
	mov dl,#0xD
	br Uncaught_Exception

   ;; INT instruction: get the number
   V86_INT:
	STOP(ds,0xB8000,23,0,'I,0x1F)
	inc word [esp+V86INT_EIP]
	mov dl,[ebx+1]
	cmp dl,#V86_SERVER_INT
	jnz V86_INT_n

   ;; Is it the SERVER being called ?
	STOP(ds,0xB8000,23,0,'S,0x1F)
	cmp ebx,#rmaddr(V86_REQUEST)	; (!) this address should be
      V86_REQUEST_Adjust:		; corrected so as to point to the
					; actual RM part
	beq V86_SERVER

   ;; Do interrupt for V86 task
   V86_INT_n:
	STOP(ds,0xB8000,23,0,'n,0x1F)
	; beware: overflow possible !!!
	sub word [esp+V86INT_ESP],6
	mov ax,[esp+V86INT_EFLAGS]
	mov [esi-2],ax
	mov ax,[esp+V86INT_CS]
	mov [esi-4],ax
	mov ax,[esp+V86INT_EIP]
	mov [esi-6],ax
	and byte [esp+V86INT_EFLAGS+1],#0xFC ; disable interrupts
	movzx edx,dl
	mov eax,[edx*4]	; interrupt vector
	mov [esp+V86INT_EIP],ax
	shr eax,16
	mov [esp+V86INT_CS],ax
	STOP(ds,0xB8000,22,0,'e,0x1F)
   ;; Return to interrupted V86 program
   V86_AGAIN:
	zzzzz({
	  sub esi,6
	  call ShowWZone ; RM stack
	  mov esi,ebx
	  call ShowBZone ; code pointer
	  mov ebx,esp
	  call DumpV86INT ; V86INT
	  z__z
	})
	STOP(ds,0xB8000,21,0,'A,0x1F)
	popad
	STOP(ds,0xB8000,21,0,'B,0x1F)
	iretd


V86INT_to_TSS:
;;Input:
;; [esi]: V86INT structure
;; [edi]: TSS32 structure
;;Output:
;;; Initialize the TSS structure
	pushing({ad,fd},{
   ;; Set registers immediately given:
	call Init_TSS0
	irp(X,{	EIP,EFLAGS,
		EAX,ECX,EDX,EBX,
		ESP,EBP,ESI,EDI,
		ES,CS,SS,DS,FS,GS },
	      {	ttmov([edi+TSS32_{}X],[esi+V86INT_{}X],eax)
	})})
	mov [edi+TSS32_ESP_CPL0],esp
	ret



;;;;;;;;;;;;;;;;; Calling Real-Mode Interrupt Handlers ;;;;;;;;;;;;;;;;;;;;

Make_V86INT:
;;Input:
;; mutual exclusion on RM stack usage for Current_V86INT assumed.
;;Output:
;; makes a V86INT structure from current gregisters and segments and
;; stack from the current V86INT.
;;TODO:
;; there should be a linked list of current V86INTs on heap.
	mov esi,[pmaddr(Current_V86INT)+sizeof_PUSHAD]
	sub esp,#sizeof_V86INT-sizeof_PUSHAD
	mov edi,esp
	pushad
	mov ecx,#(sizeof_V86INT-sizeof_PUSHAD)/4
	rep
	movsd
	ret




Call_INT_R86:
;;Input:
;; cl: interrupt number
;; [ebx]: V86INT structure
;;Output:
;; real-mode interrupt handler called in R86 mode
;; and parameter struct updated
	pushad
	pushm(  #FLAG_IOPL3|FLAG_SET,
		#pmaddr(Call_R86))
	jmp Call_X86_INT


Call_INT_V86:
;;Input:
;; cl: interrupt number
;; [ebx]: V86INT structure
;;Output:
;; real-mode interrupt handler called in V86 mode
;; and parameter struct updated
	pushad
	pushm(  #FLAG_VM|FLAG_IOPL3|FLAG_SET,
		#pmaddr(Call_V86))
	jmp Call_X86_INT ;(!) must follow


Call_INT_X86:
;;Input:
;; cl: interrupt number
;; [ebx]: V86INT structure
;;Output:
;; real-mode interrupt handler called in current X86 mode
;; and parameter struct updated
	pushad
	pushm(  #FLAG_VM|FLAG_IOPL3|FLAG_SET,
		#pmaddr(Call_X86))
	;jmp Call_X86_INT ;(!) must follow


Call_X86_INT:
;;Input:
;; [ebx]: offset of V86INT structure
;; cl: interrupt number
;; fs: X86 memory (typically ALLMEMDS)
;; [esp+8]: PUSHAD_struct, then return address
;; [esp+4]: Call_X86 address
;; [esp]: value for EFLAGS
;;Output:
;; can scramble all registers (but esp)
	STOP(fs,0xB8000,41,24,'C,0x0C)
   ;; Get and Adjust V86 stack
	SegOfsToLinear(edi,ax,[ebx+V86INT_ESP],[ebx+V86INT_SS],-10,update)
	STOP(fs,0xB8000,42,24,'a,0x0C)
	STOP(fs,0xB8000,43,24,'l,0x0C)
   ;; "Push" old CS:IP address on V86 stack
	mov eax,[ebx+V86INT_CS-2]
	mov ax,[ebx+V86INT_EIP]
	seg fs
	mov [edi+6],eax
	STOP(fs,0xB8000,44,24,'l,0x0C)
   ;; "Push" interrupt return address to X86_RETURN on V86 stack
	mov eax,[pmaddr(RM_Part_seg)-2]
	mov ax,#rmaddr(X86_RETURN)
	seg fs
	mov [edi],eax
	seg fs
	mov word [edi+4],#FLAG_IOPL3
	STOP(fs,0xB8000,45,24,'_,0x0C)
   ;; Set Target CS:IP to interrupt routine from normal real-mode IDT
	movzx ecx,cl
	shl ecx,2
	seg fs
	mov eax,[ecx]
	mov [ebx+V86INT_EIP],ax
	shr eax,16
	mov [ebx+V86INT_CS],ax
	STOP(fs,0xB8000,46,24,'X,0x0C)
   ;; Set EFLAGS
	pop eax
	xchg eax,[ebx+V86INT_EFLAGS]
	push eax
	STOP(fs,0xB8000,47,24,'8,0x0C)
   ;; There we are !
	zzzzz({call DumpV86INT})
	STOP(fs,0xB8000,48,24,'6,0x0C)
	call [esp+4]
	STOP(fs,0xB8000,49,24,'_,0x0C)
   ;; Restore EFLAGS
	pop [ebx+V86INT_EFLAGS]
	STOP(fs,0xB8000,50,24,'I,0x0C)
   ;; Pop obsolete Call_X86 argument
	pop eax
	STOP(fs,0xB8000,51,24,'N,0x0C)
   ;; Done
	STOP(fs,0xB8000,52,24,'T,0x0C)
	popad
	ret



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; X86 SERVER ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

X86_SERVER:
;;; Service manager for REQUESTs from Real-Mode to Protected-Mode.
;;Input:
;; [ebx]: V86INT of the requester
;; [esi]: linear address of the stack of the requester,
;;	  with REQUEST as first argument.
;; [esp+4]: AGAIN: RM continuation to return into RM
;; [esp]: RETURN: PM continuation in case the RM requests a return
;; Because of the way we do it, request numbers begin at 1, not 0 !
X86RQ_RETURN		=	1
X86RQ_CALL		=	2
X86RQ_PM_INT		=	3
X86RQ_RM_INT		=	4	; Intercepted RM interrupt

	STOP(fs,0xB8000,5,23,'X,0x29)
   ;; Get REQUEST number
	movzx eax, word [esi]
	STOP(fs,0xB8000,6,23,'R,0x29)
   irp({Service},
	{RETURN, CALL, PM_INT, RM_INT},{
	dec eax
	STOP(fs,0xB8000,7,23,'0+X86RQ_{}Service,0x29)
	jz X86S_{}Service
   })
  ;; Bad Request:
  ;; let's simplify things and just return !
	STOP(fs,0xB8000,7,23,'B,0x29)

X86S_RETURN:
	STOP(fs,0xB8000,8,23,'R,0x29)
	ret 4

X86S_CALL:
;;Input:
;; [ebx]: V86INT of the requester
;; [esi]: linear address of the stack of the requester,
;;	  with X86RQ_CALL as first argument.
;; [esp+4]: RM continuation to return into RM
;; [esp]: something to discard
;;Output:
	STOP(fs,0xB8000,8,23,'C,0x29)
	pop eax
	jmp [esi+2]
	;.....

X86S_PM_INT:
	STOP(fs,0xB8000,8,23,'P,0x29)
	pop eax
	mov al,[esi+2]
	mov [pmaddr(DO_X86S_PM_INT)+1],al
	jmpi pmaddr(DO_X86S_PM_INT),KL0CS ; flush instruction pipe
  DO_X86S_PM_INT:
	.byte 0xCD,0 ; interrupt instruction
	ret

X86S_RM_INT:
;;; Do everything needed to manage intercepted real-mode interrupt.
	STOP(fs,0xB8000,8,23,'I,0x29)
	;.....
	ret 4



;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Initialization ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


Init_RM_IDT:
;;Output:
;;; Initialize normal RM IDT so that usual IRQ handlers are invoked
;;; despite the 8259 PIC have been reprogrammed.
;; eax,ecx,edi scrambled
	mov edi,[pmaddr(PhysicalMem)]
	add edi,#IRQ_INT_BASE*4
	mov eax,[pmaddr(RM_Part_seg-2)]
	mov ax,#rmaddr(RM_IRQ)
	mov ecx,#16
	cld
  Init_RM_IDT_0:
	stosd
	add eax,#RM_IRQ_len
	loop Init_RM_IDT_0
	ret



;;;;;;;;;;;;;;;;;;;;;;; X86 Interface for the LLL ;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Interface from PM/LLL to V86 mode

; (!) we assume ds,es,ss point to KL0DS
; debugging assumes fs points to ALLMEMDS


DefWord({Call_INT_V86})
	Enter
	PopData(ecx)
	call Call_INT_V86
	Next

DefWord({Call_V86})
	Enter
	call Call_V86
	Next

;;; Interface from PM/LLL to R86 mode

DefWord({Call_INT_R86})
	Enter
	PopData(ecx)
	call Call_INT_R86
	Next

DefWord({Call_R86})
	Enter
	call Call_R86
	Next

})

})dnl --- end of xrv86.m4
dnl
