/* Simple installation tool */

/*
* We should eventually:
* - provide automatic recognition of current configuration
* - integrate it to the more generic OTOP sources
*/



/**************************** Include files ********************************/

#include <stdio.h>  /* for sscanf () */
#include <fcntl.h>  /* for open () */
#include <unistd.h> /* for truncate (), lseek () */
#include <errno.h>  /* for error number */
#include <stdlib.h> /* for exit() */
#include <string.h> /* for strcmp () */
#include <assert.h>
#include <sys/mman.h>
#include <sys/ipc.h>


/*********************** OS-dependent definitions **************************/

#if defined(linux)
/* We have GNU libc and more */
#  include <bytesex.h>
#  define I_GOT_MEMMEM
#  define I_GOT_STRTOUL
#  define MY_MMAP_ARG (MAP_FILE|MAP_FIXED|MAP_PRIVATE)
#elif defined(sun) || defined(m68k)
/* SunOS *sucks* */
#  define __BYTE_ORDER 4321
#  define MY_MMAP_ARG (MAP_FIXED|MAP_PRIVATE)
#else
/*#  error Unknown architecture.*/
#endif




#if   __BYTE_ORDER==1234
#  undef	SWAP_BYTES
#elif __BYTE_ORDER==4321
#  define	SWAP_BYTES
#else
#  error Unknown bytesex !
#endif


#ifndef I_GOT_MEMMEM
static void *memmem(const void *needle, size_t needlelen,
              const void *haystack, size_t haystacklen)
{
  for (;haystacklen>=needlelen;haystack=(char*)haystack+1,haystacklen--)
     if(!memcmp(haystack,needle,needlelen)) return haystack ;
  return NULL ;
}
#endif



/************************* Type/Size Definitions ****************************/


typedef unsigned long		dword ;
typedef unsigned short		word ;
typedef unsigned char		byte ;

long PAGE_SIZE ;
long PAGE_MASK ;

#define SECTSIZE 512
#define BSIZE 16384
unsigned char BUFFER [BSIZE] ;		/* buffer for I/O */



/**************************** Error checking ********************************/

#define errprintf(x...)	    fprintf(stderr,x)
#define CHECK(x,good,msg)   (((x)==(good))?0:(errprintf(msg "\n"),exit(100+errno)))
#define CHECKN(x,bad,msg)   (((x)!=(bad))?0:(errprintf(msg "\n"),exit(100+errno)))

#undef DEBUG
#ifdef DEBUG
int __DBG__=1 ;
#  define DBG(x)	(__DBG__?(x):0)
#else
#  define DBG(x)
#endif






/******** CPU-Endianness independent endian-aware memory/disk access *******/

#define putLEbyte putbyte
#define putBEbyte putbyte
#define getLEbyte getbyte
#define getBEbyte getbyte
inline void putbyte(void*x,byte y) {
  *((byte*)x)=y ;
}
inline byte getbyte(void*x) {
  return *((byte*)x) ;
}

inline void putLEword(void*x,word y) {
  ((byte*)x)[0]=(byte)y ;
  ((byte*)x)[1]=(byte)(y>>8) ;
}
inline void putLEdword(void*x,dword y) {
  ((byte*)x)[0]=(byte)y ;
  ((byte*)x)[1]=(byte)(y>>8) ;
  ((byte*)x)[2]=(byte)(y>>16) ;
  ((byte*)x)[3]=(byte)(y>>24) ;
}
inline word getLEword(void*x) {
  return  ((byte*)x)[0] | (((byte*)x)[1]<<8) ;
}
inline dword getLEdword(void*x) {
  return    ((byte*)x)[0]      | (((byte*)x)[1]<<8) 
         | (((byte*)x)[2]<<16) | (((byte*)x)[3]<<24) ;
}

inline void putBEword(void*x,word y) {
  ((byte*)x)[1]=(byte)y ;
  ((byte*)x)[0]=(byte)(y>>8) ;
}
inline void putBEdword(void*x,dword y) {
  ((byte*)x)[3]=(byte)y ;
  ((byte*)x)[2]=(byte)(y>>8) ;
  ((byte*)x)[1]=(byte)(y>>16) ;
  ((byte*)x)[0]=(byte)(y>>24) ;
}
inline word getBEword(void*x) {
  return  ((byte*)x)[1] | (((byte*)x)[0]<<8) ;
}
inline dword getBEdword(void*x) {
  return    ((byte*)x)[3]      | (((byte*)x)[2]<<8) 
         | (((byte*)x)[1]<<16) | (((byte*)x)[0]<<24) ;
}

#define writeLEbyte writebyte
#define readLEbyte readbyte
#define writeBEbyte writebyte
#define readBEbyte readbyte

void writebyte(int fd,long pos,byte b) {
  CHECKN(  lseek (fd,pos,SEEK_SET), -1, "Couldn't position on byte (w)."  );
  CHECKN(  write (fd,&b,1), -1, "Couldn't write byte."  );
}
byte readbyte(int fd,long pos) {
  byte x ;
  CHECKN(  lseek (fd,pos,SEEK_SET), -1, "Couldn't position on byte (r)."  );
  CHECKN(  read (fd,&x,1), -1, "Couldn't read byte."  );
  return x ;
}

void writeLEword(int fd,long pos,word w) {
  byte x[2] ;
  putLEword(x,w) ;
  CHECKN(  lseek (fd,pos,SEEK_SET), -1, "Couldn't position on word (w)."  );
  CHECKN(  write (fd,&x,2), -1, "Couldn't write word."  );
}
void writeLEdword(int fd,long pos,dword l) {
  byte x[4] ;
  putLEdword(x,l) ;
  CHECKN(  lseek (fd,pos,SEEK_SET), -1, "Couldn't position on dword (w)."  );
  CHECKN(  write (fd,&x,4), -1, "Couldn't write dword."  );
}

word readLEword(int fd,long pos) {
  byte x[2] ;
  CHECKN(  lseek (fd,pos,SEEK_SET), -1, "Couldn't position on word (r)."  );
  CHECKN(  read (fd,&x,2), -1, "Couldn't read word."  );
  return getLEword(x) ;
}
dword readLEdword(int fd,long pos) {
  byte x[4] ;
  CHECKN(  lseek (fd,pos,SEEK_SET), -1, "Couldn't position on dword (r)."  );
  CHECKN(  read (fd,&x,4), -1, "Couldn't read dword."  );
  return getLEdword(x) ;
}



/**** OLD STUFF

#ifdef SWAP_BYTES
#  define gooddword(x) (swapbytes(x))
#  define filedword(x) (swapbytes(x))
#  define goodword(x) (wswapbytes(x))
#  define fileword(x) (wswapbytes(x))
#else
#  define gooddword(x) (x)
#  define filedword(x) (x)
#  define goodword(x) (x)
#  define fileword(x) (x)
#endif
inline void reversebytes(char*x,int i) {
  int j ;
  char t ;
  for(j=0;j+j<i;j++) {
    t=x[j] ;
    x[j]=x[i-j] ;
    x[i-j]=t ;
  }
}
inline dword swapbytes(dword x) {
  x = (x<<16) | (x>>16) ;
  x = ((x&0x00FF00FF)<<8) | (x&0xFF00FF00>>8) ;
  return x ;
}
inline word wswapbytes(word x) {
   x = (x<<8) | (x>>8) ;
   return x ;
}

******/





/************************** Simple subroutines *****************************/


long filesize(int fd)
{
  long l ;
  CHECKN(l = lseek(fd,0,SEEK_END), -1, "Couldn't get file size.") ;
  return l ;
}

dword getenvint (char*A,dword def) {
  char *X=getenv(A) ;
  return X ? strtoul(X,NULL,0) : def ;
}
long getenvchar (char*A,char def) {
  char *X=getenv(A) ;
  return X ? *X : def ;
}




/*************************** Simple Check sum ******************************/

dword LEchksum(int fd,long beg,long len)
{
  long Z,i,j,L ;
  dword S=0 ;


  CHECKN(  lseek (fd,beg,SEEK_SET),-1,"Couldn't reset file."  );

  for(i=0;i+BSIZE<len;i+=BSIZE) {
    CHECK(  read (fd,BUFFER,BSIZE), BSIZE, "Error filling buffer."  );
    for(j=0;j<BSIZE;j+=4) {
      S += getLEdword(BUFFER+j) ;
    }
  }


  L=len-i ;
  CHECK(  read (fd,BUFFER,L), L, "Error filling last buffer."  );
  memset(BUFFER+L,BSIZE-L,0) ;
    for(j=0;j<L;j+=4) {
      S += getLEdword(BUFFER+j) ;
    }

  return S ;
}



#ifdef DEBUG
dword partLEchksum(int fd,long beg,long len)
{
  long Z,i,j,k,L ;
  dword S=0 ;


  CHECKN(  lseek (fd,beg,SEEK_SET),-1,"Couldn't reset file."  );

  for(i=0;i+BSIZE<len;i+=BSIZE) {
    CHECK(  read (fd,BUFFER,BSIZE), BSIZE, "Error filling buffer."  );
    for(k=0;k<BSIZE;k+=SECTSIZE) {
      for(j=0;j<SECTSIZE;j+=4) {
	S += getLEdword(BUFFER+k+j) ;
      }
      errprintf("Partial Checksum 0x%08lX.\n",S) ;
    }
  }

  L=len-i ;
  CHECK(  read (fd,BUFFER,L), L, "Error filling last buffer."  );
  memset(BUFFER+L,BSIZE-L,0) ;

    for(k=0;k<L;k+=SECTSIZE) {
      for(j=0;j<SECTSIZE;j+=4) {
	S += getLEdword(BUFFER+k+j) ;
      }
      errprintf("Partial checksum 0x%08lX.\n",S) ;
    }

  return S ;
}
#endif



/************************ Main program functions *****************************/


void toggle_i386_option (int fd) {

/* To use on biosapi.b to toggle options without recompiling the image...
 *
 */
  char *B,*O,*OSTR,*Z,*Z0 ;
  long L0,L,OLEN,OPT ;

  CHECK(readLEdword(fd,56), 0,	"checksum field should be zero.") ;
  L0 = filesize(fd) ;
  CHECKN(  Z0 = malloc(L0+3*PAGE_SIZE), 0, "Couldn't get memory zone."  ) ;
  Z = (void*)( ((long)(Z0+2*PAGE_SIZE)+PAGE_SIZE-1)&PAGE_MASK ) ;
  L = L0 ;

  DBG(  errprintf("Z0=%08lX Z=%08lX L=%ld.\n",Z0,Z,L)  ) ;


  CHECKN( (long)( B = mmap(Z, L,
		 PROT_READ, MY_MMAP_ARG,
		 fd, 0)), -1, "Couldn't mmap()."  ) ;

  CHECKN(  OSTR = getenv("OPTION_SIG"), 0, "I need an option name."  ) ;
  CHECKN(  OLEN = strlen(OSTR), 0, "Uh ??"  ) ;
  CHECKN(  OPT = getenvint("OPTION_NUM", 1234), 1234,
	 	"I need an option number."  ) ;
  CHECK( (OPT+1)/2, 1, "Option number must be 1 or 2." ) ;

  O = memmem(OSTR,OLEN,B,L) ;

  DBG(  errprintf("B=%08lX L=%ld OSTR=%s OLEN=%d O=%08lX.\n",B,L,OSTR,OLEN,O)  ) ;

  CHECKN(  O , 0, "Option not found."  ) ;
  writeLEdword(fd,O-Z-13,getLEdword(O+4*(-3+OPT))) ;
}



void adjust_i386_image (int fd) {

/* This routine adjusts compiler output into a bootable system image.
 *
 * Arguments come from environment:
 * SYSSIZE: size without bootrec+loader
 * ...
 */

  dword BOOTRECSIZE	= getenvint("BOOTRECSIZE",0) ;
  dword LOADERSIZE	= getenvint("LOADERSIZE",0) ;
  dword SYSSIZE		= getenvint("SYSSIZE",0) ;

  char LOADMODE		= getenvchar("LOADMODE",'C') ;

  dword APIOFS		= BOOTRECSIZE+LOADERSIZE ;
  dword TOTALSIZE	= BOOTRECSIZE+LOADERSIZE+SYSSIZE ;
  dword SYSSIZE_B	= SECTSIZE*SYSSIZE ;
  dword APIOFS_B	= SECTSIZE*APIOFS ;
  dword TOTALSIZE_B	= SECTSIZE*TOTALSIZE ;

  dword checksum ;

  CHECK(BOOTRECSIZE,		1,	"Boot sector over 512 bytes !") ;
  CHECK(readLEword(fd,510),	0xAA55,	"Invalid boot sector signature.") ;
  CHECK(readLEdword(fd,APIOFS_B+56), 0,	"checksum field should be zero.") ;


  writeLEword(fd,0x00E,		TOTALSIZE) ;	/* reserved sectors */
  writebyte(fd,0x024,		LOADMODE) ;	/* Load Mode */
  writeLEword(fd,0x025,		APIOFS) ;	/* where loading shall begin */
  writebyte(fd,0x02A,		LOADERSIZE) ;	/* loader size for bootrec */
  writebyte(fd,0x1F1,		LOADERSIZE) ;	/* loader size for bootrec */
  writeLEword(fd,0x1F4,		SYSSIZE_B>>4) ; /* syssize in para for LILO */
  writeLEdword(fd,APIOFS_B+32,	SYSSIZE_B) ;	/* syssize for loader */

  checksum = LEchksum(fd,APIOFS_B,SYSSIZE_B) ;
  writeLEdword(fd,APIOFS_B+56,	-checksum) ;    /* checksum for loader */

  DBG(  errprintf("Checksum is 0x%08lX.\n", checksum)  ) ;
  DBG(  partLEchksum(fd,APIOFS_B,SYSSIZE_B)  ) ;


  errprintf("Total image size is %ld bytes (%ld sectors).\n",
	 TOTALSIZE_B,TOTALSIZE) ;
}


void report_error(void) {
  errprintf("You shouldn't run this program manually.\n") ;
  exit(2) ;
}



/**************************** Argument parser ********************************/

typedef void (*action_t)(int fd) ;

#define namefun(x)  {#x,(action_t)x}

struct tasks {
  char *name ;
  action_t fun ;
} task [] =
{ namefun(adjust_i386_image),
  namefun(toggle_i386_option),
  {NULL,(action_t)report_error}
} ;


int find_task(char*n) {
   int i=0 ;
   while(task[i].name) {
     if(!strcmp(task[i].name,n)) { break ;}
     else i++ ;
   }
   return i ;
}


int main(int argc, char **argv)
{
  int fd ;

  PAGE_SIZE = getpagesize() ;
  PAGE_MASK = ~(PAGE_SIZE-1) ;

  if (argc!=3) report_error() ;

  CHECKN(fd = open(argv[2],O_RDWR),0, "Couldn't open argument file.") ;

  (*(task[find_task(argv[1])].fun))(fd) ;

  DBG(  errprintf("Done with %s %s.\n",argv[1],argv[2])  );

  return 0 ;
}
